__doc__=r"""
:py:mod:`known/store.py`
"""
import os

__all__ = ['Node', 'Store']

def isValidNodeName(k:str): return ((not (k.startswith('__') and k.endswith('__'))) and k.isidentifier())

class Node:
    r""" Node Object - represents a directory tree.
    Each node has:
    * ```__path__``` which specifies absolute path of Node's root
    * ```__files__``` a list of file names in the directory
    * sub-directories are set as attributes of Node type
    
    .. note:: calling a Node object with no args returns it full path, if any argument is provides, uses ```os.join``` to append it

    .. note:: Not to be used directly, these are attributes on ```Store``` objects.
    """

    def __init__(self, path, members=None) -> None:
        if members:
            for k,v in members.items(): 
                if isValidNodeName(k): setattr(self, k, v)
        self.__path__ = path #<--- path will overwrite if present in members
        self.__ref__()

    def __ref__(self): self.__files__ = [ f for f in os.listdir(self.__path__) if os.path.isfile(os.path.join( self.__path__,f)) ]

    def __call__(self, file=None, relative=False): 
        p = os.path.join(self.__path__, file) if file else self.__path__
        if relative: p = os.path.relpath( p )
        return p

    def __repr__(self) -> str:return self.__str__()
    def __str__(self) -> str:  return f'Node @ [{self()}] :: {len([k for k in self.__dict__ if isValidNodeName(k)])} Dirs, {len(self.__files__)} Files'

    
class Store:
    r""" Abstracts a directory tree as collection of Nodes

    :param cwd_node: (bool) if True, adds ```os.getcwd()``` to Store
    :param auto_nodes: (list of str) adds nodes with their base-name as identifiers
    :param named_nodes: (dict of name:path) adds nodes with specified name as identifiers

    .. note:: Directories (nodes) with valid names are added as attributes. 
        Invalid directories and their sub-directories are ignored.
        A valid name should be an identifier and must not be dunder like i.e., does not start and end with `__` chars.
        
    """
    @staticmethod
    def build_tree_recursive(root_path):
        fL = os.listdir(root_path)
        node=Node(root_path)
        for f in fL:
            fp = os.path.join(root_path, f)
            if os.path.isdir(fp):
                k,v = __class__.build_tree_recursive(fp)
                setattr( node, k, v )
        return os.path.basename(root_path), node

    @staticmethod
    def call_tree_recursive(root, method):
        _ = getattr(root, method)() if isinstance(method, str) else method(root)
        for k,v in root.__dict__.items():
            if isValidNodeName(k): __class__.call_tree_recursive(v, method)
    
    def __repr__(self) -> str: return self.__str__()
    def __str__(self) -> str: return f'Store: [{len(self.__roots__)}] : {self.__roots__.keys()}'

    def show(self):
        for k,v in self.__roots__.items():
            print(f'\n[KEY: {k}]')
            __class__.call_tree_recursive(v,print)

    def __init__(self, cwd_node:bool, *auto_nodes, **named_nodes) -> None:
        self.__roots__ = {}
        self.add_auto_nodes(*auto_nodes)
        self.add_named_nodes(**named_nodes)
        if cwd_node: self.add_node(os.getcwd())

    def refresh(self, node):
        if node in self.__roots__:
            res = self.add_node(self.__roots__[node].__path__)
            if not res: 
                del self.__roots__[node]
                delattr (self, node )

    def add_node(self, directory, create=False):
        r""" add a single node from a directory """
        if not os.path.isdir(directory):
            if not create: 
                print(f'[.] path @ {directory} is not a directory, skipping...')
                return False
    
        rp = os.path.abspath(directory)
        rn = os.path.basename(rp)
        if isValidNodeName(rn): 
            if create: os.makedirs(rp, exist_ok=True)
            n, r = __class__.build_tree_recursive(rp)
            if hasattr(self, n): print(f'[!] [{n}] already exsists in directories, will be overwritten!')
            setattr(self, n, r)
            self.__roots__[n]=r
            return True
        else:
            print(f'[!] [{rn}] is invalid name, skipping...')
            return False

    def add_named_nodes(self, **directories):
        r""" add multiple nodes from a dict  """
        for n,directory in directories.items():
            if not os.path.isdir(directory):
                print(f'[.] path @ {directory} is not a directory, skipping...')
                continue
            rp = os.path.abspath(directory)
            if isValidNodeName(n): 
                _, r = __class__.build_tree_recursive(rp)
                if hasattr(self, n): print(f'[!] [{n}] already exsists in directories, will be overwritten!')
                setattr(self, n, r)
                self.__roots__[n]=r
            else:
                print(f'[!] [{n}] is invalid name, skipping...')

    def add_auto_nodes(self, *directories):
        r""" add multiple nodes from a list  """
        for directory in directories:
            if not os.path.isdir(directory):
                print(f'[.] path @ {directory} is not a directory, skipping...')
                continue
            rp = os.path.abspath(directory)
            rn = os.path.basename(rp)
            if isValidNodeName(rn): 
                n, r = __class__.build_tree_recursive(rp)
                if hasattr(self, n): print(f'[!] [{n}] already exsists in directories, will be overwritten!')
                setattr(self, n, r)
                self.__roots__[n]=r
            else:
                print(f'[!] [{rn}] is invalid name, skipping...')




#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
__doc__=r"""
:py:mod:`known/basic.py`
"""
#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
__all__ = [
    'now', 'numel', 'arange', 'save_json', 'load_json', 'save_pickle', 'load_pickle',
     
    'Symbols', 'Verbose', 'BaseConvert', 'IndexedDict', 'Remap', 
    #==============================
    
    'onehot', 'onecold', 'dict_sort', 
]
#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
import datetime
from typing import Any, Union, Iterable #, BinaryIO, cast, Dict, Optional, Type, Tuple, IO
import numpy as np
from numpy import ndarray
from math import floor, log, ceil
import json, pickle
from collections import UserDict
#import pickle, pathlib, io
#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

def now(year:bool=True, month:bool=True, day:bool=True, 
        hour:bool=True, minute:bool=True, second:bool=True, mirco:bool=True, 
        start:str='', sep:str='', end:str='') -> str:
    r""" Unique Identifier - useful in generating unique identifiers based on current timestamp. 
    Helpful in generating unique filenames based on timestamps. 
    
    .. seealso::
        :func:`~known.basic.Verbose.strU`
    """
    form = []
    if year:    form.append("%Y")
    if month:   form.append("%m")
    if day:     form.append("%d")
    if hour:    form.append("%H")
    if minute:  form.append("%M")
    if second:  form.append("%S")
    if mirco:   form.append("%f")
    assert (form), 'format should not be empty!'
    return (start + datetime.datetime.strftime(datetime.datetime.now(), sep.join(form)) + end)

def numel(shape:Iterable) -> int: 
    r""" Returns the number of elements in an array of given shape. """
    return np.prod(np.array(shape))

def arange(shape:Iterable, start:int=0, step:int=1, dtype=None) -> ndarray: 
    r""" Similar to ``np.arange`` but reshapes the array to given shape. """
    return np.arange(start=start, stop=start+step*numel(shape), step=step, dtype=dtype).reshape(shape)

def save_json(o:Any, path:str, **kwargs) -> None:
    r""" save object to json file """
    with open(path, 'w') as f: json.dump(o, f, **kwargs)

def load_json(path:str) -> Any:
    r""" load json file to object """
    with open(path, 'r') as f: o = json.load(f)
    return o

def save_pickle(o:Any, path:str,**kwargs):
    r""" save object to pickle file """
    with open(path, 'wb') as f: pickle.dump(o, f,**kwargs)

def load_pickle(path:str):
    r""" load pickle file to object """
    with open(path, 'rb') as f: o = pickle.load(f)
    return o

def onehot(total:int, index:int, **kwargs):
    res = np.zeros(total, **kwargs)
    res[index]=1
    return res

def onecold(total:int, index:int, **kwargs):
    res = np.ones(total, **kwargs)
    res[index]=0
    return res

def dict_sort(D:dict, assending:bool=True, fin=lambda x:x, fmid=lambda x:x, fout=lambda x:x, return_dict=True):
    r""" if D is like :: dict[str, numbers], then sorts  the numbers and keys
        
    1. ```fin``` is applied on each number and it is appended to a list
    2. ```fmid``` is applied on the list
    3. create np.array from the output of fmid
    4. ```fout``` is applied on the ndarray
    5. argsort is applied on ndarray
    
    returns keys, values and indices in sorted order
    """
    # sorts the values and keys into lists
    K,V = [], []
    for k,v in D.items():
        K.append(k)
        V.append(fin(v))
    K=np.array(K)
    V=fout(np.array(fmid(V)))
    S = np.argsort(V)
    if assending:
        ks = K[S]
        vs = V[S]
        ss = S 
    else:
        ks = K[S][::-1]
        vs = V[S][::-1]
        ss = S[::-1]
    
    return {k:v for k,v in zip(ks,vs)} if return_dict else (ks,vs,ss)






#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

class Symbols:
    r"""
    contains some special symbols described in the table below

    .. list-table:: 
        :widths: 5 3 5 3
        :header-rows: 1

        * - Name
          - Symbol
          - Name
          - Symbol
        * - CORRECT
          - ✓
          - INCORRECT
          - ✗
        * - ALPHA
          - α
          - BETA
          - β
        * - GAMMA
          - γ
          - DELTA
          - δ
        * - EPSILON
          - ε
          - ZETA
          - ζ
        * - ETA
          - η
          - THETA
          - θ
        * - KAPPA
          - κ
          - LAMBDA
          - λ
        * - MU
          - μ 
          - XI
          - ξ
        * - PI
          - π
          - ROH
          - ρ
        * - SIGMA
          - σ
          - PHI
          - φ
        * - PSI
          - Ψ
          - TAU
          - τ
        * - OMEGA
          - Ω
          - TRI
          - Δ
    """
    
    CORRECT =       '✓'
    INCORRECT =     '✗'
    ALPHA =         'α'
    BETA =          'β'
    GAMMA =         'γ'
    DELTA =         'δ'
    EPSILON =       'ε'
    ZETA =          'ζ'
    ETA =           'η'
    THETA =         'θ'
    KAPPA =         'κ'
    LAMBDA =        'λ'
    MU =            'μ' 
    XI =            'ξ'
    PI =            'π'
    ROH =           'ρ'
    SIGMA =         'σ'
    PHI =           'φ'
    PSI =           'Ψ'
    TAU =           'τ'
    OMEGA =         'Ω'
    TRI =           'Δ'

#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

class Verbose:
    r""" Contains shorthand helper functions for printing outputs and representing objects as strings.

    .. note::
        This class contains only static methods.
    """
    DEFAULT_DATE_FORMAT = ["%Y","%m","%d","%H","%M","%S","%f"]
    r""" Default date format for :func:`~known.basic.Verbose.strU` """

    DASHED_LINE = "=-=-=-=-==-=-=-=-="
    DOCSTR_FORM = lambda x: f'\t!docstr:\n! - - - - - - - - - - - - - - - - -\n{x}\n- - - - - - - - - - - - - - - - - !'

    @staticmethod
    def strN(s:str, n:int) -> str:  
        r""" Repeates a string n-times """
        return ''.join([s for _ in range(n)])

    @staticmethod
    def _recP_(a, level, index, pindex, tabchar='\t', show_dim=False):
        # helper function for recP - do not use directly
        if index<0: index=''
        dimstr = ('* ' if level<1 else f'*{level-1} ') if show_dim else ''
        pindex = f'{pindex}{index}'
        if len(a.shape)==0:
            print(f'{__class__.strN(tabchar, level)}[ {dimstr}@{pindex}\t {a} ]') 
        else:
            print(f'{__class__.strN(tabchar, level)}[ {dimstr}@{pindex} #{a.shape[0]}')
            for i,s in enumerate(a):
                __class__._recP_(s, level+1, i, pindex, tabchar, show_dim)
            print(f'{__class__.strN(tabchar, level)}]')

    @staticmethod
    def recP(arr:Iterable, show_dim:bool=False) -> None: 
        r"""
        Recursive Print - print an iterable recursively with added indentation.

        :param arr:         any iterable with ``shape`` property.
        :param show_dim:    if `True`, prints the dimension at the start of each item
        """
        __class__._recP_(arr, 0, -1, '', '\t', show_dim)
    
    @staticmethod
    def strA(arr:Iterable, start:str="", sep:str="|", end:str="") -> str:
        r"""
        String Array - returns a string representation of an iterable for printing.
        
        :param arr:     input iterable
        :param start:   string prefix
        :param sep:     item seperator
        :param end:     string postfix
        """
        res=start
        for a in arr: res += (str(a) + sep)
        return res + end

    @staticmethod
    def strD(arr:Iterable, sep:str="\n", cep:str=":\n", caption:str="") -> str:
        r"""
        String Dict - returns a string representation of a dict object for printing.
        
        :param arr:     input dict
        :param sep:     item seperator
        :param cep:     key-value seperator
        :param caption: heading at the top
        """
        res=f"=-=-=-=-==-=-=-=-={sep}DICT #[{len(arr)}] : {caption}{sep}{__class__.DASHED_LINE}{sep}"
        for k,v in arr.items(): res+=str(k) + cep + str(v) + sep
        return f"{res}{__class__.DASHED_LINE}{sep}"

    @staticmethod
    def strU(form:Union[None, Iterable[str]], start:str='', sep:str='', end:str='') -> str:
        r""" 
        String UID - returns a formated string of current timestamp.

        :param form: the format of timestamp, If `None`, uses the default :data:`~known.basic.Verbose.DEFAULT_DATE_FORMAT`.
            Can be selected from a sub-set of ``["%Y","%m","%d","%H","%M","%S","%f"]``.
            
        :param start: UID prefix
        :param sep: UID seperator
        :param end: UID postfix

        .. seealso::
            :func:`~known.basic.uid`
        """
        if not form: form = __class__.DEFAULT_DATE_FORMAT
        return start + datetime.datetime.strftime(datetime.datetime.now(), sep.join(form)) + end

    @staticmethod
    def show(x:Any, cep:str='\t\t:', sw:str='__', ew:str='__') -> None:
        r"""
        Show Object - describes members of an object using the ``dir`` call.

        :param x:       the object to be described
        :param cep:     the name-value seperator
        :param sw:      argument for ``startswith`` to check in member name
        :param ew:      argument for ``endswith`` to check in member name

        .. note:: ``string.startswith`` and ``string.endswith`` checks are performed on each member of the object 
            and only matching member are displayed. This is usually done to prevent showing dunder members.
        
        .. seealso::
            :func:`~known.basic.Verbose.showX`
        """
        for d in dir(x):
            if not (d.startswith(sw) or d.endswith(ew)):
                v = ""
                try:
                    v = getattr(x, d)
                except:
                    v='?'
                print(d, cep, v)

    @staticmethod
    def showX(x:Any, cep:str='\t\t:') -> None:
        """ Show Object (Xtended) - describes members of an object using the ``dir`` call.

        :param x:       the object to be described
        :param cep:     the name-value seperator

        .. note:: This is the same as :func:`~known.basic.Verbose.show` but skips ``startswith`` and ``endswith`` checks,
            all members are shown including dunder members.

        .. seealso::
            :func:`~known.basic.Verbose.show`
        """
        for d in dir(x):
            v = ""
            try:
                v = getattr(x, d)
            except:
                v='?'
            print(d, cep, v)

    @staticmethod
    def dir(x:Any, doc=False, filter:str='', sew=('__','__')):
        """ Calls ```dir``` on given argument and lists the name and types of non-dunder members.

        :param filter: csv string of types to filter out like `type,function,module`, keep blank for no filter
        :param doc: shows docstring ```__doc``` 
            If ```doc``` is True, show all member's ```__doc__```.
            If ```doc``` is False, does not show any ```__doc__```. 
            If ```doc``` is a string, show ```__doc__``` of specific types only given by csv string.

        :param sew: 2-Tuple (start:str, end:str) - excludes member names that start and end with specific chars, 
            used to exclude dunder methods by default
        """
        #if self_doc: print( f'{type(x)}\n{x.__doc__}\n' )
        if sew: sw, ew = f'{sew[0]}', f'{sew[1]}'
        doc_is_specified = (isinstance(doc, str) and bool(doc))
        if doc_is_specified: doc_match =[ t for t in doc.replace(' ','').split(',') if t ]
        if filter: filter_match =[ t for t in filter.replace(' ','').split(',') if t ]
        counter=1
        for k in dir(x):
            if sew:
                if (k.startswith(sw) and k.endswith(ew)): continue
            m = getattr(x,k)
            n = str(type(m)).split("'")[1]
            if filter:
                if not (n in filter_match):  continue
            s = f'[{counter}] {k} :: {n}'#.encode('utf-16')

            if doc:
                if doc_is_specified:
                    if n in doc_match: 
                        d = __class__.DOCSTR_FORM(m.__doc__)
                    else:
                        d=''
                else:
                    d = __class__.DOCSTR_FORM(m.__doc__)
            else:
                d = ''
            counter+=1
            print(f'{s}{d}')


    @staticmethod
    def info(x:Any, show_object:bool=False):
        r""" Shows the `type`, `length` and `shape` of an object and optionally shows the object as well.

        :param x:           the object to get info about
        :param show_object: if `True`, prints the object itself

        .. note:: This is used to check output of some functions without having to print the full output
            which may take up a lot of console space. Useful when the object are of nested types.

        .. seealso::
            :func:`~known.basic.Verbose.infos`
        """
        print(f'type: {type(x)}')
        if hasattr(x, '__len__'):
            print(f'len: {len(x)}')
        if hasattr(x, 'shape'):
            print(f'shape: {x.shape}')
        if show_object:
            print(f'object:\n{x}')

    @staticmethod
    def infos(X:Iterable, show_object=False):
        r""" Shows the `type`, `length` and `shape` of each object in an iterable 
        and optionally shows the object as well.

        :param x:           the object to get info about
        :param show_object: if `True`, prints the object itself

        .. seealso::
            :func:`~known.basic.Verbose.info`
        """
        for t,x in enumerate(X):
            print(f'[# {t}]')
            __class__.info(x, show_object=show_object)

#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

class Remap:
    r""" 
    Provides a mapping between ranges, works with scalars, ndarrays and tensors.

    :param Input_Range:     *FROM* range for ``i2o`` call, *TO* range for ``o2i`` call
    :param Output_Range:    *TO* range for ``i2o`` call, *FROM* range for ``o2i`` call

    .. note::
        * :func:`~known.basic.REMAP.i2o`: maps an input within `Input_Range` to output within `Output_Range`
        * :func:`~known.basic.REMAP.o2i`: maps an input within `Output_Range` to output within `Input_Range`

    Examples::

        >>> mapper = REMAP(Input_Range=(-1, 1), Output_Range=(0,10))
        >>> x = np.linspace(mapper.input_low, mapper.input_high, num=5)
        >>> y = np.linspace(mapper.output_low, mapper.output_high, num=5)

        >>> yt = mapper.i2o(x)  #<--- should be y
        >>> xt = mapper.o2i(y) #<----- should be x
        >>> xE = np.sum(np.abs(yt - y)) #<----- should be 0
        >>> yE = np.sum(np.abs(xt - x)) #<----- should be 0
        >>> print(f'{xE}, {yE}')
        0, 0
    """

    def __init__(self, Input_Range:tuple, Output_Range:tuple) -> None:
        r"""
        :param Input_Range:     `from` range for ``i2o`` call, `to` range for ``o2i`` call
        :param Output_Range:    `to` range for ``i2o`` call, `from` range for ``o2i`` call
        """
        self.set_input_range(Input_Range)
        self.set_output_range(Output_Range)

    def set_input_range(self, Range:tuple) -> None:
        r""" set the input range """
        self.input_low, self.input_high = Range
        self.input_delta = self.input_high - self.input_low

    def set_output_range(self, Range:tuple) -> None:
        r""" set the output range """
        self.output_low, self.output_high = Range
        self.output_delta = self.output_high - self.output_low

    def backward(self, X):
        r""" maps ``X`` from ``Output_Range`` to ``Input_Range`` """
        return ((X - self.output_low)*self.input_delta/self.output_delta) + self.input_low

    def forward(self, X):
        r""" maps ``X`` from ``Input_Range`` to ``Output_Range`` """
        return ((X - self.input_low)*self.output_delta/self.input_delta) + self.output_low

    def __call__(self, X, backward=False):
        return self.backward(X) if backward else self.forward(X)
    
    def swap_range(self):
        Input_Range, Output_Range = (self.output_low, self.output_high), (self.input_low, self.input_high)
        self.set_input_range(Input_Range)
        self.set_output_range(Output_Range)

#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

class IndexedDict(UserDict):
    r""" Implements an Indexed dict where values can be addressed using both index(int) and keys(str) """

    def __init__(self, **members) -> None:
        self.names = []
        super().__init__(*[], **members)
    
    def keys(self): return enumerate(self.names, 0) # for i,k in self.keys()

    def items(self): return enumerate(self.data.items(), 0) # for i,(k,v) in self.items()

    def __len__(self): return len(self.data)

    def __getitem__(self, name): 
        if isinstance(name, int): name = self.names[name]
        if name in self.data: 
            return self.data[name]
        else:
            raise KeyError(name)

    def __setitem__(self, name, item): 
        if isinstance(name, int): name = self.names[name]
        if name not in self.data: self.names.append(name)
        self.data[name] = item

    def __delitem__(self, name): 
        index = None
        if isinstance(name, int):  
            index = name
            name = self.names[name]
        if name in self.data: 
            del self.names[self.names.index(name) if index is None else index]
            del self.data[name]

    def __iter__(self): return iter(self.names)

    def __contains__(self, name): return name in self.data

    # Now, add the methods in dicts but not in MutableMapping

    def __repr__(self) -> str:
        return f'{__class__} :: {len(self)} Members'
    
    def __str__(self) -> str:
        items = ''
        for i,k in enumerate(self):
            items += f'[{i}] \t {k} : {self[i]}\n'
        return f'{__class__} :: {len(self)} Members\n{items}'
    
    def __copy__(self):
        inst = self.__class__.__new__(self.__class__)
        inst.__dict__.update(self.__dict__)
        # Create a copy and avoid triggering descriptors
        inst.__dict__["data"] = self.__dict__["data"].copy()
        inst.__dict__["names"] = self.__dict__["names"].copy()
        return inst

#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

class BaseConvert:
    r""" Number System Conversion """


    @staticmethod
    def ndigs(num:int, base:int) -> int:
        r""" 
        Returns the number of digits required to represent a base-10 number in the given base.

        :param num:     base-10 number to be represented
        :param base:    base-n number system
        """
        return 1 + (0 if num==0 else floor(log(num, base)))

    @staticmethod
    def int2base(num:int, base:int, digs:int) -> list:
        r""" 
        Convert base-10 integer to a base-n list of fixed no. of digits 

        :param num:     base-10 number to be represented
        :param base:    base-n number system
        :param digs:    no of digits in the output

        :returns:       represented number as a list of ordinals in base-n number system

        .. seealso::
            :func:`~known.basic.base2int`
        """
        if not digs: digs=__class__.ndigs(num, base)
        res = [ 0 for _ in range(digs) ]
        q = num
        for i in range(digs): # <-- do not use enumerate plz
            res[i]=q%base
            q = floor(q/base)
        return res

    @staticmethod
    def base2int(num:Iterable, base:int) -> int:
        """ 
        Convert an iterbale of digits in base-n system to base-10 integer

        :param num:     iterable of base-n digits
        :param base:    base-n number system

        :returns:       represented number as a integer in base-10 number system

        .. seealso::
            :func:`~known.basic.int2base`
        """
        res = 0
        for i,n in enumerate(num): res+=(base**i)*n
        return int(res)


    SYM_BIN = { f'{i}':i for i in range(2) }
    SYM_OCT = { f'{i}':i for i in range(8) }
    SYM_DEC = { f'{i}':i for i in range(10) }
    SYM_HEX = {**SYM_DEC , **{ s:(i+10) for i,s in enumerate(('A', 'B', 'C', 'D', 'E', 'F'))}}
    
    @staticmethod
    def n_syms(n): return { f'{i}':i for i in range(n) }

    @staticmethod
    def to_base_10(syms:dict, num:str):
        b = len(syms)
        l = [ syms[n] for n in num[::-1] ]
        return __class__.base2int(l, b)

    @staticmethod
    def from_base_10(syms:dict, num:int, joiner=''):
        base = len(syms)
        print(f'----{num=} {type(num)}, {base=}, {type(base)}')
        ndig = 1 + (0 if num==0 else floor(log(num, base))) # __class__.ndigs(num, base)
        ss = tuple(syms.keys())
        S = [ ss[i]  for i in __class__.int2base(num, base, ndig) ]
        return joiner.join(S[::-1])


    @staticmethod
    def int2hex(num:int, joiner=''): return __class__.from_base_10(__class__.SYM_HEX, num, joiner)
        



class BaseConverter:
    r""" Number System Conversion 
    
    A number is abstract concept that has many representations using sets of symbols

    A base-n number system uses a set of n digits to represent any number
    This is called the representation of the number

    Given one representation, we only need to convert to another

    """

    DIGITS_DTYPE =  np.uint32
    @staticmethod
    def convert(digits, base_from, base_to, reversed=True):
        r""" convers from one base to another 
        
        :param digits:      iterable of digits in base ```base_from```. NOTE: digits are Natural Numbers starting at 0. base 'b' will have digits between [0, b-1]
        :param base_from:   int - the base to convert from
        :param base_to:     int - the base to convert to
        :param reversed:    bool - if True, digits are assumed in reverse (human readable left to right)
                            e.g. if reversed is True then binary digits iterable [1,0,0] will represent [4] in decimal otherwise it will represent [1] in decimal
        """

        digits_from =  np.array(digits, dtype=__class__.DIGITS_DTYPE) # convert to int data-type
        if reversed: digits_from = digits_from[::-1]
        ndigits_from = len(digits_from)
        mult_from = np.array([base_from**i for i in range(ndigits_from)], dtype=__class__.DIGITS_DTYPE)
        repr_from = np.dot(digits_from , mult_from)

        #ndc = base_from**ndigits_from
        ndigits_to = ceil(log(repr_from,base_to))
        digits_to =  np.zeros((ndigits_to,), dtype=__class__.DIGITS_DTYPE)
        n = int(repr_from)
        for d in range(ndigits_to):
            digits_to[d] = n%base_to
            n=n//base_to
            #n = (n-digits_to[d])/base_to

        #mult_to = np.array([base_to**i for i in range(ndigits_to)])
        #repr_to = np.dot(digits_to , mult_to)
        #assert repr_to==repr_from
        if reversed: digits_to = digits_to[::-1]
        return tuple(digits_to)


    @staticmethod
    def int2base(num:int, base:int, digs:int) -> list:
        r""" 
        Convert base-10 integer to a base-n list of fixed no. of digits 

        :param num:     base-10 number to be represented
        :param base:    base-n number system
        :param digs:    no of digits in the output

        :returns:       represented number as a list of ordinals in base-n number system

        .. seealso::
            :func:`~known.basic.base2int`
        """
        
        ndigits = digs if digs else ceil(log(num,base)) 
        digits =  np.zeros((ndigits,), dtype=__class__.DIGITS_DTYPE)
        n = num
        for d in range(ndigits):
            digits[d] = n%base
            n=n//base
        return digits

    @staticmethod
    def base2int(num:Iterable, base:int) -> int:
        """ 
        Convert an iterbale of digits in base-n system to base-10 integer

        :param num:     iterable of base-n digits
        :param base:    base-n number system

        :returns:       represented number as a integer in base-10 number system

        .. seealso::
            :func:`~known.basic.int2base`
        """
        res = 0
        for i,n in enumerate(num): res+=(base**i)*n
        return int(res)


    SYM_BIN = { f'{i}':i for i in range(2) }
    SYM_OCT = { f'{i}':i for i in range(8) }
    SYM_DEC = { f'{i}':i for i in range(10) }
    SYM_HEX = {**SYM_DEC , **{ s:(i+10) for i,s in enumerate(('A', 'B', 'C', 'D', 'E', 'F'))}}
    
    @staticmethod
    def n_syms(n): return { f'{i}':i for i in range(n) }

    @staticmethod
    def to_base_10(syms:dict, num:str):
        b = len(syms)
        l = [ syms[n] for n in num[::-1] ]
        return __class__.base2int(l, b)

    @staticmethod
    def from_base_10(syms:dict, num:int, joiner=''):
        base = len(syms)
        print(f'----{num=} {type(num)}, {base=}, {type(base)}')
        ndig = 1 + (0 if num==0 else floor(log(num, base))) # __class__.ndigs(num, base)
        ss = tuple(syms.keys())
        S = [ ss[i]  for i in __class__.int2base(num, base, ndig) ]
        return joiner.join(S[::-1])


    @staticmethod
    def int2hex(num:int, joiner=''): return __class__.from_base_10(__class__.SYM_HEX, num, joiner)
        




# known
__doc__=r"""
**known** is a collection of reusable python code. Use pip to install **known**

.. code-block:: console

   $ pip install known
   
The package is frequently updated by adding new functionality, make sure to have the latest version. To check version use

.. code-block:: console

   >>> import known
   >>> known.__version___

:py:mod:`known/__init__.py`
"""
__version__ = '0.0.7'
#print(f'known.{__version__}')


# from .basic import *
# from .imgu import *
# from .mailer import *
# from .store import *




#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
__doc__=r"""
:py:mod:`known/mailer.py`
"""
#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
import os, platform, smtplib, mimetypes, pickle #, json # imghdr
from email.message import EmailMessage
from zipfile import ZipFile

__all__=['Mail']


class Mail:
    r""" Use a g-mail account to send mail. 

    .. warning:: This requires saving your login credentials on machine in pickle format. 
        You should enable 2-factor-auth in gmail and generate an app-password instead of using your gmail password.
        Visit (https://myaccount.google.com/apppasswords) to generate app-password.
        Usually, these type of emails are treated as spam by google, so they must be marked 'not spam' at least once.
        It is recomended to create a seperate gmail account for sending mails.

    :param login_path: the path to login file

    A login file should be created first using ```save_login(path)``` (static) method. This will ask you to enter username and password
    which will be saved at the supplied path. The same path must be used for ```login_path``` argument.

    The usename and password is read from the file and used to login to gmail server every time ```send()``` or ```send_mail()``` is called.

    > use the ```Mail.send()``` static method to send emails.

    """
    
    DEFAULT_CTYPE = 'application/octet-stream'  

    @staticmethod
    def global_alias(prefix=''): return f'{prefix}{os.getlogin()} @ {platform.node()}:{platform.system()}.{platform.release()}'

    @staticmethod
    def zip_files(zip_path:str, files, **kwargs):
        r""" zips all (only files) in the list of file paths and saves at 'zip_path' """
        zipped = 0
        if not zip_path.lower().endswith('.zip'): zip_path = f'{zip_path}.zip'
        with ZipFile(zip_path, 'w', **kwargs) as zip_object:
            for path in files:
                if not os.path.isfile(path): continue
                zip_object.write(f'{path}')
                zipped+=1
        return zipped, zip_path

    @staticmethod
    def get_all_file_paths(directory):
        r""" recursively list all files in a folder """
        file_paths = []
        # crawling through directory and subdirectories
        for root, directories, files in os.walk(directory):
            for filename in files:
                # join the two strings in order to form the full filepath.
                filepath = os.path.join(root, filename)
                file_paths.append(filepath)
        return file_paths   

    @staticmethod
    def zip_folders(zip_path:str, folders, **kwargs):  
        r""" zip multiple folders into a single zip file """    
        if isinstance(folders, str): folders= [f'{folders}']

        if not zip_path : zip_path = f'{folders[0]}.zip'
        if not zip_path.lower().endswith('.zip'): zip_path = f'{zip_path}.zip'  
        all_files = []
        for folder in folders: all_files.extend(__class__.get_all_file_paths(folder))
        return __class__.zip_files(f'{zip_path}', all_files, **kwargs)
    
    @staticmethod
    def zip_folder(folder:str, **kwargs):
        r""" zip a single folder with the same zip file name """     
        return  __class__.zip_files(f'{folder}.zip', __class__.get_all_file_paths(folder),  **kwargs)


    def __init__(self, login_path:str, signature=None, verbose=False) -> None: 
        self.verbose=verbose
        self.login_path=login_path
        if not os.path.isfile(login_path): print(f'{__class__} :: path @ [{login_path}] seems invalid, make sure to first create a login file using the "save_login" method.')
        self.signature = self.__class__.global_alias(prefix='\n') if signature is None else signature 
        
    def __call__(self, subject, rx, cc, bcc, content, attached = None): 
        r""" sends an e-mail message 

        :param subject: (str)
        :param rx: Recivers, csv string for 'To' field
        :param cc:  CarbonCopy, csv string for 'Cc' field
        :param bcc:  Blind CarbonCopy, csv string for 'Bcc' field
        :param content: lines to go inside msg body
        :param attached: (List of 2-Tuple) attachements for the email
        
        """
        
        self.__class__.send_mail(lambda : self.__class__.load_login(self.login_path),
        self.__class__.compose_mail(
            subject = f'{subject}', 
            rx = rx, cc = cc, bcc = bcc, 
            content = content + self.signature, 
            attached = attached, 
            verbose=self.verbose,)
            )

    @staticmethod
    def get_mime_types(files):
        r""" gets mimetype info all files in a list """
        if isinstance(files, str): files=[f'{files}']
        res = []
        for path in files:
            if not os.path.isfile(path): continue
            ctype, encoding = mimetypes.guess_type(path)
            if ctype is None or encoding is not None: ctype = __class__.DEFAULT_CTYPE  
            maintype, subtype = ctype.split('/', 1)
            res.append( (path, maintype, subtype) )
        return res

    @staticmethod
    def compose_mail( subject:str, rx:str, cc:str, bcc:str, content:str, attached, verbose=True):
        r""" compose an e-mail msg to send later
        
        :param subject: subject
        :param rx: csv recipent email address
        :param cc: csv cc email address
        :param content: main content
        :param attached: list of attached files - is a 2-tupe (attachment_type, (args...) )

        # attach all files in the list :: ('', ('file1.xyz', 'file2.xyz'))
        # zip all the files in the list :: ('zipname.zip', '(file1.xyz', 'file2.xyz'))
        """
        
        msg = EmailMessage()

        # set subject
        msg['Subject'] = f'{subject}'
        if verbose: print(f'SUBJECT: {subject}')

        # set to
        msg['To'] = rx
        if verbose: print(f'TO: {rx}')

        if cc: msg['Cc'] = cc
        if verbose: print(f'CC: {cc}')

        if bcc: msg['Bcc'] = bcc
        if verbose: print(f'BCC: {bcc}')

        # set content
        msg.set_content(content)
        if verbose: print(f'MESSAGE: #[{len(content)}] chars.')

        default_attached = []

        attached = [] if attached is None else attached
        for (attach_type, attach_args) in attached:
            if verbose: print(f' ... processing attachement :: {attach_type} :: {attach_args}')

            all_files = []
            for path in attach_args:
                if os.path.isdir(path):
                    all_files.extend(__class__.get_all_file_paths(path))
                elif os.path.isfile(path):
                    all_files.append(path)
                else:
                    if verbose: print(f'[!] Invalid Path :: {path}, skipped...')

            if not attach_type:  # attach individually
                default_attached.extend(__class__.get_mime_types(all_files))
            else: # make zip
                zipped, zip_path=__class__.zip_files(attach_type, all_files)
                if verbose: print(f'\t --> zipped {zipped} items @ {zip_path} ')
                if zipped>0:
                    default_attached.extend(__class__.get_mime_types(zip_path))
                else:
                    if verbose: print(f'[!] [{zip_path}] is empty, will not be attched!' )
                    try:
                        os.remove(zip_path)
                        if verbose: print(f'[!] [{zip_path}] was removed.' )
                    except:
                        if verbose: print(f'[!] [{zip_path}] could not be removed.' ) 
                

        # set attached ( name, main_type, sub_type), if sub_type is none, auto-infers using imghdr
        for file_name,main_type,sub_type in default_attached:
            if verbose: print(f'[+] Attaching file [{main_type}/{sub_type}] :: [{file_name}]')
            with open (file_name, 'rb') as f: 
                file_data = f.read()
            msg.add_attachment(file_data, maintype=main_type, subtype=sub_type, filename=os.path.basename(file_name))

        return msg

    @staticmethod
    def send_mail(login, msg, verbose=True):
        r""" send a msg using smtp.gmail.com:587 with provided credentials """
        username, password = login()
        if verbose: print(f'[*] Sending Email from {username}')
        msg['From'] = f'{username}' # set from
        with smtplib.SMTP('smtp.gmail.com', 587) as smpt:
            smpt.starttls()
            smpt.login(username, password)
            smpt.ehlo()
            smpt.send_message(msg)
        if verbose: print(f'[*] Sent!')


    @staticmethod
    def send(username:str, password:str, subject:str, rx:str, cc:str, bcc:str, content:str, attached, verbose=True):
        login = lambda: (username, password)
        msg = __class__.compose_mail(subject, rx, cc, bcc, content, attached, verbose)
        __class__.send_mail(login, msg, verbose)

    @staticmethod
    def save_pickle(obj, path:str,**kwargs):
        r""" save object to pickle file """
        with open(path, 'wb') as f: pickle.dump(obj, f,**kwargs)

    @staticmethod
    def load_pickle(path:str):
        r""" load pickle file to object """
        with open(path, 'rb') as f: o = pickle.load(f)
        return o

    @staticmethod
    def str2bytes(s:str, encoding:str='raw_unicode_escape')->list: return [b for b in bytes(s, encoding)]

    @staticmethod
    def bytes2str(s, encoding:str='raw_unicode_escape')->str: return bytes.decode(bytes(s), encoding)

    @staticmethod
    def save_login(path):
        r""" save your login credentials as pickle """
        __class__.save_pickle((  __class__.str2bytes(input('Enter Username')), __class__.str2bytes(input('Enter Password'))  ), path)
    
    @staticmethod
    def load_login(path):
        r""" load your login credentials from json """
        login = __class__.load_pickle(path)
        return __class__.bytes2str(login[0]), __class__.bytes2str(login[1])
    #--------------------------------------------------


from setuptools import setup

setup(
    name =                      'known',
    version =                   '0.0.7',
    url =                       'https://github.com/Nelson-iitp/known',
    author =                    'Nelson.S',
    author_email =              'mail.nelsonsharma@gmail.com',
    description =               'Module :: known',
    packages =                  ['known'],
    classifiers=                ['License :: OSI Approved :: MIT License'],
    package_dir =               { '' : 'module'},
    install_requires =          [],
    include_package_data =      True
)   


__doc__=r"""
:py:mod:`known/imgu.py`
"""
#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
__all__ = [
    'Pix', 'graphfromimage',
]#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
#import datetime, os
import numpy as np
from numpy import ndarray
import matplotlib.pyplot as plt
import cv2 # pip install opencv-python

from .basic import BaseConvert



CHANNELS = 4
DTYPE = np.uint8
CHANNEL_BRGA = (0, 1, 2, 3)
CHANNEL_RBGA = (1, 2, 0, 3)


class Pix(object):
    r""" abstracts a 4-channel (brga) image """

    def __init__(self, h, w, create=True) -> None:
        self.h, self.w = int(h), int(w)
        if create: self.i = np.zeros((self.h, self.w, CHANNELS), dtype=DTYPE) 
    
    @property
    def RBGA(self): return self.i[:, :, CHANNEL_RBGA]

    def plot_on(self, ax, grid): 
        ax.imshow(self.RBGA)
        if grid:
            xtick, ytick = np.arange(self.w), np.arange(self.h)
            ax.set_xticks(xtick-0.5)
            ax.set_xticklabels(xtick)
            ax.set_yticks(ytick-0.5)
            ax.set_yticklabels(ytick)
            ax.grid(axis='both')

    def clear(self, channel=None): 
        if channel is None:     self.i[:,:,:]       =0 # clear all channels
        else:                   self.i[:,:,channel] =0 # clear specified channel

    def get_color_at(self, row, col, normalize=False):
        b,g,r,a = (self.i[row, col, :]/255 if normalize else self.i[row, col, :])
        return (r, g, b, a)
    
    def set_color_at(self, row:int, col:int, rgba:tuple, normalize=False): 
        if normalize: rgba = [int(x*255) for x in rgba]
        r,g,b,a = rgba
        self.i[row, col, :] = (b, r, g, a) 
    

    def set_hex_at(self, row:int, col:int, hex:str):
        if hex.startswith('#'): hex = hex[1:]
        hex = hex.upper()
        lenhex = len(hex)
        assert lenhex==6 or lenhex==8, f'expecting 6 or 8 chars but got {lenhex} :: {hex}'
        if lenhex==6: hex = 'FF' + hex # max alpha
        B,G,R,A = tuple(BaseConvert.int2base(num=BaseConvert.to_base_10(BaseConvert.SYM_HEX, hex), base=256, digs=4))
        return self.set_color_at(row,col,(R,G,B,A))
    

    def get_hex_at(self, row:int, col:int):
        r,g,b,a = self.get_color_at(row, col, normalize=False)
        int_code = BaseConvert.base2int(num=(b,g,r,a), base=256)
        print(f'{int_code=}, {type(int_code)}')
        hex_code = BaseConvert.int2hex(int_code,joiner='')
        print(f'{hex_code=}')
        return hex_code
    
    @staticmethod
    def save(pix, path):  
        return cv2.imwrite(path, pix.i)

    @staticmethod
    def load(path): 
        img =  cv2.imread(path, cv2.IMREAD_UNCHANGED)
        assert img.ndim==3, f'expecting 3-D array but got {img.ndim}-D'
        assert img.shape[-1]== CHANNELS, f'must be argb image with {CHANNELS} channels but got {img.shape[-1]} channels'
        h, w, _ = img.shape
        pix = __class__(h, w, False)
        pix.i = img.astype(DTYPE)
        return pix

    @staticmethod
    def plot(pix, ratio=0.75, grid=True):
        fig, ax = plt.subplots(1,1, figsize=(pix.w*ratio, pix.h*ratio))
        pix.plot_on(ax, grid)
        plt.show()

    def show_color_at(self, row, col, fs=1):
        rgba = self.get_color_at(row,col, normalize=False)
        hexc = self.get_hex_at(row, col)
        color = self.get_color_at(row, col, normalize=True)
        if not isinstance(fs, tuple): fs=(fs,fs)
        plt.figure(figsize=fs)
        plt.yticks([], [])
        plt.title(f'{rgba=}')
        plt.bar([f'{hexc}'], [1], color=color)
        plt.show()
        plt.close()


def graphfromimage(img_path:str, pixel_choice:str='first', dtype=None) -> ndarray:
    r""" 
    Covert an image to an array (1-Dimensional)

    :param img_path:        path of input image 
    :param pixel_choice:    choose from ``[ 'first', 'last', 'mid', 'mean' ]``

    :returns: 1-D numpy array containing the data points

    .. note:: 
        * This is used to generate synthetic data in 1-Dimension. 
            The width of the image is the number of points (x-axis),
            while the height of the image is the range of data points, choosen based on their index along y-axis.
    
        * The provided image is opened in grayscale mode.
            All the *black pixels* are considered as data points.
            If there are multiple black points in a column then ``pixel_choice`` argument specifies which pixel to choose.

        * Requires ``opencv-python``

            Input image should be readable using ``cv2.imread``.
            Use ``pip install opencv-python`` to install ``cv2`` package
    """
    img= cv2.imread(img_path, 0)
    imgmax = img.shape[1]-1
    j = img*0
    j[np.where(img==0)]=1
    pixel_choice = pixel_choice.lower()
    pixel_choice_dict = {
        'first':    (lambda ai: ai[0]),
        'last':     (lambda ai: ai[-1]),
        'mid':      (lambda ai: ai[int(len(ai)/2)]),
        'mean':     (lambda ai: np.mean(ai))
    }
    px = pixel_choice_dict[pixel_choice]
    if dtype is None: dtype=np.float_
    return np.array([ imgmax-px(np.where(j[:,i]==1)[0]) for i in range(j.shape[1]) ], dtype=dtype)


